/*
 * Created by diablos on 2007-12-02 17:51:40 +0100
 * All Rights Reserved
 */

#ifndef _PAVEOPERATORS_H_
#define _PAVEOPERATORS_H_

#import <AppKit/AppKit.h>

#include <lua50/lua.h>
#include <lua50/lauxlib.h>
#include <lua50/lualib.h>

// some useful definitions first

static inline void
push_point (lua_State *L, NSPoint p)
{
  lua_newtable (L);

  lua_pushstring (L, "x");
  lua_pushnumber (L, p.x);
  lua_settable (L, -3);

  lua_pushstring (L, "y");
  lua_pushnumber (L, p.y);
  lua_settable (L, -3);
}

static inline void
push_size (lua_State *L, NSSize s)
{
  lua_newtable (L);

  lua_pushstring (L, "width");
  lua_pushnumber (L, s.width);
  lua_settable (L, -3);

  lua_pushstring (L, "height");
  lua_pushnumber (L, s.height);
  lua_settable (L, -3);
}

static inline void
push_rect (lua_State *L, NSRect r)
{
  lua_newtable (L);

  lua_pushstring (L, "origin");
  push_point (L, r.origin);
  lua_settable (L, -3);

  lua_pushstring (L, "size");
  push_size (L, r.size);
  lua_settable (L, -3);
}

// generic non-argument operator
#define OP(__op_name__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  PS ## __op_name__ (); \
  return 0; \
}

// numeric set-operators
#define S_NUM_OP1(__op_name__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  double a = luaL_checknumber (L, 1); \
  PS ## __op_name__ (a); \
  return 0; \
}

#define S_NUM_OP2(__op_name__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  double a = luaL_checknumber (L, 1), \
         b = luaL_checknumber (L, 2); \
  PS ## __op_name__ (a, b); \
  return 0; \
}

#define S_NUM_OP3(__op_name__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  double a = luaL_checknumber (L, 1), \
         b = luaL_checknumber (L, 2), \
         c = luaL_checknumber (L, 3); \
  PS ## __op_name__ (a, b, c); \
  return 0; \
}

#define S_NUM_OP4(__op_name__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  double a = luaL_checknumber (L, 1), \
         b = luaL_checknumber (L, 2), \
         c = luaL_checknumber (L, 3), \
         d = luaL_checknumber (L, 4); \
  PS ## __op_name__ (a, b, c, d); \
  return 0; \
}

#define S_NUM_OP5(__op_name__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  double a = luaL_checknumber (L, 1), \
         b = luaL_checknumber (L, 2), \
         c = luaL_checknumber (L, 3), \
         d = luaL_checknumber (L, 4), \
         e = luaL_checknumber (L, 5); \
  PS ## __op_name__ (a, b, c, d, e); \
  return 0; \
}

#define S_NUM_OP6(__op_name__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  double a = luaL_checknumber (L, 1), \
         b = luaL_checknumber (L, 2), \
         c = luaL_checknumber (L, 3), \
         d = luaL_checknumber (L, 4), \
         e = luaL_checknumber (L, 5), \
         f = luaL_checknumber (L, 6); \
  PS ## __op_name__ (a, b, c, d, e, f); \
  return 0; \
}

// numeric get-operators
#define G_NUM_OP1(__op_name__, __type__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  __type__ a; \
  PS ## __op_name__ (&a); \
  lua_pushnumber (L, (double) a); \
  return 1; \
}

#define G_NUM_OP2(__op_name__, __type__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  __type__ a, b; \
  PS ## __op_name__ (&a, &b); \
  lua_pushnumber (L, (double) a); \
  lua_pushnumber (L, (double) b); \
  return 2; \
}

#define G_NUM_OP3(__op_name__, __type__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  __type__ a, b, c; \
  PS ## __op_name__ (&a, &b, &c); \
  lua_pushnumber (L, (double) a); \
  lua_pushnumber (L, (double) b); \
  lua_pushnumber (L, (double) c); \
  return 3; \
}

#define G_NUM_OP4(__op_name__, __type__) \
static int pave_ ## __op_name__ (lua_State *L) \
{ \
  __type__ a, b, c, d; \
  PS ## __op_name__ (&a, &b, &c, &d); \
  lua_pushnumber (L, (double) a); \
  lua_pushnumber (L, (double) b); \
  lua_pushnumber (L, (double) c); \
  lua_pushnumber (L, (double) d); \
  return 4; \
}


// after that, defining PostScript operators is quick and easy

// color operations
G_NUM_OP1   (currentalpha, float)
G_NUM_OP4   (currentcmykcolor, float)
G_NUM_OP1   (currentgray, float)
G_NUM_OP3   (currenthsbcolor, float)
G_NUM_OP3   (currentrgbcolor, float)

S_NUM_OP1   (setalpha)
S_NUM_OP1   (setgray)
S_NUM_OP3   (setrgbcolor)
S_NUM_OP3   (sethsbcolor)
S_NUM_OP4   (setcmykcolor)


// Gstate handling
OP          (gsave)
OP          (grestore)

// Gstate operations
G_NUM_OP1   (currentflat, float)
G_NUM_OP1   (currentlinecap, int)
G_NUM_OP1   (currentlinejoin, int)
G_NUM_OP1   (currentlinewidth, float)
G_NUM_OP1   (currentmiterlimit, float)
G_NUM_OP2   (currentpoint, float)
G_NUM_OP1   (currentstrokeadjust, int)

S_NUM_OP1   (setflat)
S_NUM_OP2   (sethalftonephase)
S_NUM_OP1   (setlinecap)
S_NUM_OP1   (setlinejoin)
S_NUM_OP1   (setlinewidth)
S_NUM_OP1   (setmiterlimit)
S_NUM_OP1   (setstrokeadjust)

// Matrix operations
static int pave_concat (lua_State *L)
{
  float matrix[6];

  matrix[0] = luaL_checknumber (L, 1);
  matrix[1] = luaL_checknumber (L, 2);
  matrix[2] = luaL_checknumber (L, 3);
  matrix[3] = luaL_checknumber (L, 4);
  matrix[4] = luaL_checknumber (L, 5);
  matrix[5] = luaL_checknumber (L, 6);

  PSconcat (matrix);

  return 0;
}

OP          (initmatrix)
S_NUM_OP1   (rotate)
S_NUM_OP2   (scale)
S_NUM_OP2   (translate)

// Paint operations
S_NUM_OP5   (arc)
S_NUM_OP5   (arcn)
S_NUM_OP5   (arct)
OP          (clip)
OP          (closepath)
S_NUM_OP6   (curveto)
OP          (eoclip)
OP          (eofill)
OP          (fill)
OP          (flattenpath)
OP          (initclip)
S_NUM_OP2   (lineto)
S_NUM_OP2   (moveto)
OP          (newpath)
G_NUM_OP4   (pathbbox, float)
S_NUM_OP6   (rcurveto)
S_NUM_OP4   (rectclip)
S_NUM_OP4   (rectfill)
S_NUM_OP4   (rectstroke)
OP          (reversepath)
S_NUM_OP2   (rlineto)
S_NUM_OP2   (rmoveto)
OP          (stroke)

#undef OP

#undef S_NUM_OP1
#undef S_NUM_OP2
#undef S_NUM_OP3
#undef S_NUM_OP4
#undef S_NUM_OP5
#undef S_NUM_OP6

#undef G_NUM_OP1
#undef G_NUM_OP2
#undef G_NUM_OP3
#undef G_NUM_OP4

static int
pave_availablefonts (lua_State *L)
{
  lua_newtable (L);

  NSArray *fonts = [[NSFontManager sharedFontManager]
    availableFonts];

  int i, n;
  for (i = 0, n = [fonts count]; i < n; i++)
    {
      lua_pushstring (L, [[fonts objectAtIndex: i] cString]);
      lua_rawseti (L, -2, i);
    }

  return 1;
}

static int
pave_availablefontfamilies (lua_State *L)
{
  lua_newtable (L);

  NSArray *fontFamilies = [[NSFontManager sharedFontManager]
    availableFontFamilies];

  int i, n;
  for (i = 0, n = [fontFamilies count]; i < n; i++)
    {
      lua_pushstring (L, [[fontFamilies objectAtIndex: i] cString]);
      lua_rawseti (L, -2, i);
    }

  return 1;
}

static int
pave_findfont (lua_State *L)
{
  NSString *fontName = [NSString stringWithCString: luaL_checkstring (L, 1)];
  float size = luaL_checknumber (L, 2);

  NSFont *f = [NSFont fontWithName: fontName size: size];

  if (f == nil)
    {
      lua_pushnil (L);
    }
  else
    {
      lua_pushlightuserdata (L, f);
    }

  return 1;
}

static int
pave_setfont (lua_State *L)
{
  luaL_checktype (L, 1, LUA_TLIGHTUSERDATA);
  NSFont *f = (NSFont *) lua_topointer (L, 1);

  [f set];

  return 0;
}

static int
pave_fontinfo (lua_State *L)
{
  luaL_checktype (L, 1, LUA_TLIGHTUSERDATA);
  NSFont *f = (NSFont *) lua_topointer (L, 1);

  lua_newtable (L);

  lua_pushstring (L, "boundingRect");
  push_rect (L, [f boundingRectForFont]);
  lua_settable (L, -3);

  lua_pushstring (L, "displayName");
  lua_pushstring (L, [[f displayName] cString]);
  lua_settable (L, -3);

  lua_pushstring (L, "familyName");
  lua_pushstring (L, [[f familyName] cString]);
  lua_settable (L, -3);

  lua_pushstring (L, "fontName");
  lua_pushstring (L, [[f fontName] cString]);
  lua_settable (L, -3);

  lua_pushstring (L, "encodingScheme");
  lua_pushstring (L, [[f encodingScheme] cString]);
  lua_settable (L, -3);

  lua_pushstring (L, "isFixedPitch");
  lua_pushboolean (L, [f isFixedPitch]);
  lua_settable (L, -3);

  lua_pushstring (L, "isBaseFont");
  lua_pushboolean (L, [f isBaseFont]);
  lua_settable (L, -3);

  lua_pushstring (L, "pointSize");
  lua_pushnumber (L, [f pointSize]);
  lua_settable (L, -3);

  lua_pushstring (L, "ascender");
  lua_pushnumber (L, [f ascender]);
  lua_settable (L, -3);

  lua_pushstring (L, "descender");
  lua_pushnumber (L, [f descender]);
  lua_settable (L, -3);

  lua_pushstring (L, "capHeight");
  lua_pushnumber (L, [f capHeight]);
  lua_settable (L, -3);

  lua_pushstring (L, "italicAngle");
  lua_pushnumber (L, [f italicAngle]);
  lua_settable (L, -3);

  lua_pushstring (L, "maximumAdvancement");
  push_size (L, [f maximumAdvancement]);
  lua_settable (L, -3);

  lua_pushstring (L, "minimumAdvancement");
  push_size (L, [f minimumAdvancement]);
  lua_settable (L, -3);

  lua_pushstring (L, "underlinePosition");
  lua_pushnumber (L, [f underlinePosition]);
  lua_settable (L, -3);

  lua_pushstring (L, "underlineThickness");
  lua_pushnumber (L, [f underlineThickness]);
  lua_settable (L, -3);

  lua_pushstring (L, "xHeight");
  lua_pushnumber (L, [f xHeight]);
  lua_settable (L, -3);

  lua_pushstring (L, "defaultLineHeight");
  lua_pushnumber (L, [f defaultLineHeightForFont]);
  lua_settable (L, -3);

  return 1;
}

static int
pave_widthofstring (lua_State *L)
{
  luaL_checktype (L, 1, LUA_TLIGHTUSERDATA);
  NSFont *f = (NSFont *) lua_topointer (L, 1);

  NSString *str = [NSString stringWithCString: luaL_checkstring (L, 2)];

  lua_pushnumber (L, [f widthOfString: str]);

  return 1;
}

static int
pave_show (lua_State *L)
{
  const char *str = luaL_checkstring (L, 1);

  PSshow (str);

  return 0;
}

static int
pave_charpath (lua_State *L)
{
  const char *str = luaL_checkstring (L, 1);

  luaL_checktype(L, 2, LUA_TBOOLEAN);
  int flag = lua_toboolean (L, 2);

  PScharpath (str, flag);

  return 0;
}

static int
pave_setdash (lua_State *L)
{
  luaL_checktype(L, 1, LUA_TTABLE);
  float offset = luaL_checknumber (L, 2);

  int i, n = luaL_getn (L, 1);
  float *pattern = malloc (n * sizeof (float));

  for (i = 0; i < n; i++)
    {
      lua_rawgeti (L, 1, i + 1);

      pattern[i] = luaL_checknumber (L, -1);

      if (pattern[i] == 0)
        {
          free (pattern);

          luaL_error (L, "Invalid dash length in setdash pattern at offset %d "
            "- the dash length must be a non-zero number", i);
        }

      lua_pop (L, 1);
    }

  PSsetdash (pattern, n, offset);

  free (pattern);

  return 0;
}

#endif // _PAVEOPERATORS_H_
